#!/bin/bash
set -eo pipefail

log()
{
	echo "# $* #"
}

log_err()
{
	RED='\033[0;31m'
	NC='\033[0m'

	echo -e "${RED}# $* #${NC}"
}

exec_cmd()
{
	log "Running [$@]"
	$@
	if [ "$?" -ne 0 ]; then
		log_err "Failed to run [$@]..."
		if [ -n "${CI}" ]; then
			exit 1
		else
			# Ignore errors for development environment
			log_err "Ignoring the last error ..."
		fi
	fi
}

generate_compilation_db()
{
	if [ -f "compile_commands.json" ]; then
		if jq -e '. | length == 0' compile_commands.json; then
			log_err "Compilation db empty, probably COMPILATION_FIXUP not set"
			exit 1
		fi
		log "Compilation db already exits, skip generation"
		return 0
	fi

	if [ -n "${COMPILATION_FIXUP}" ]; then
		COMPILATION_FIXUP="${COMPILATION_FIXUP/cmake /cmake -DCMAKE_EXPORT_COMPILE_COMMANDS=ON }"
		exec_cmd ${COMPILATION_FIXUP}
	fi

	if [ -f "compile_commands.json" ]; then
		if jq -e '. | length == 0' compile_commands.json; then
			log_err "Empty compilation db, probably COMPILATION_FIXUP incorrect"
			exit 3
		fi
		log "Compilation db got generated with COMPILATION_FIXUP ..."
		return 0
	fi

	exec_cmd bear -- make -C ${SOURCE_FOLDER}
	make -C ${SOURCE_FOLDER} clean

	if [ ! -f "compile_commands.json" ]; then
		log_err "Failed to generate compilation db"
		exit 2
	fi

	if jq -e '. | length == 0' compile_commands.json; then
		log_err "Empty compilation db, probably COMPILATION_FIXUP not set"
		exit 4
	fi
}

run_cppcheck_validation()
{
	mkdir -p /tmp/cppcheck
	if [ -f "compile_commands.json" ]; then
		exec_cmd cppcheck --error-exitcode=1 --addon=threadsafety --addon=cert -i ./test --inline-suppr ${CPPCHECK_OPTIONS} --project=compile_commands.json --cppcheck-build-dir=/tmp/cppcheck
	fi
}

run_cppcheck_clang_validation()
{
	mkdir -p /tmp/cppcheck
	if [ -f "compile_commands.json" ]; then
		if [ -n "${CI}" ]; then
			cppcheck --error-exitcode=1 --clang -i ./test --inline-suppr ${CPPCHECK_OPTIONS} --project=compile_commands.json --cppcheck-build-dir=/tmp/cppcheck
		else
			exec_cmd cppcheck --error-exitcode=1 --clang -i ./test --inline-suppr ${CPPCHECK_OPTIONS} --project=compile_commands.json --cppcheck-build-dir=/tmp/cppcheck
		fi
	fi
}

run_infer_analysis()
{
	cmd="infer --fail-on-issue --compilation-database compile_commands.json -o /tmp/infer ${INFER_OPTIONS}"
	if ! ${cmd}; then
		log_err "Failed to execute [$cmd]"
		cp /tmp/infer/report.txt .
	fi
}

run_flawfinder_checks()
{
    exec_cmd flawfinder --minlevel 4 --error-level=4 ${FLAWFINDER_OPTIONS} ${SOURCE_FOLDER}
}

run_cpd_check()
{
    exec_cmd /usr/local/pmd/bin/run.sh cpd --language c --exclude ./test/ ${CPD_OPTIONS} --files ${SOURCE_FOLDER}
}

validate_c_code()
{
	log "SOURCE_FOLDER: ${SOURCE_FOLDER}"
	log "COMPILATION_FIXUP: ${COMPILATION_FIXUP}"
	log "CPPCHECK_OPTIONS: ${CPPCHECK_OPTIONS}"
	log "FLAWFINDER_OPTIONS: ${FLAWFINDER_OPTIONS}"
	log "CPD_OPTIONS: ${CPD_OPTIONS:=--minimum-tokens 200}"
	log ""

	if [ -z "${SOURCE_FOLDER}" ]; then
		log "SOURCE_FOLDER not defined"
		if [ -n "${CI}" ]; then
			log "Can't proceed without SOURCE_FOLDER in pipeline"
			exit 5
		else
			log "Assuming current path '.' as SOURCE_FOLDER"
			export SOURCE_FOLDER="."
		fi
	fi

	if [ "${SOURCE_FOLDER}" == "SKIP" ]; then
		log "Skipping C code analysis"
		exit 0
	fi

	# Install pre-req if present/defined
	if [ -f "./gitlab-ci/install-dependencies.sh" ]; then
		log "Installing prerequisites for compilation"
		exec_cmd ./gitlab-ci/install-dependencies.sh
	fi

	# Run CPD checks
	run_cpd_check

	# Run flawfinder
	run_flawfinder_checks

	# Generate compilation db
	generate_compilation_db

	# TODO: Run gcc fanalyzer analysis 

	# Run cppcheck with compilation-db
	run_cppcheck_validation

	# Run clang analysis
	# run_cppcheck_clang_validation

	# Run infer analysis
	run_infer_analysis
}

validate_shell_script()
{
	log "Running optional tests"
	if [ -z "${SHELL_SRC}" ]; then
		log "Skipping ShellCheck path not defined"
		exit 0
	fi

	if [ "${SHELL_SRC}" == "SKIP" ]; then
		log "Skipping ShellChecks ..."
		exit 0
	fi

	log "Running shellcheck ${SHELLCHECK_OPTIONS} ${SHELL_SRC}"
	exec_cmd shellcheck ${SHELLCHECK_OPTIONS} ${SHELL_SRC}
}

validation_type="${1}"

case "${validation_type}" in
	c_code)
		validate_c_code "${@}"
		;;
	shell_script)
		validate_shell_script "${@}"
		;;
	*)
		echo "Unsupported validation type: ${validation_type}"
		exit 1
	;;
esac
